# coding: utf-8 -*- (Coding Format)

# ------------------------------------------------------------------------------------------------------

# PostalDataProcessingScript.py

# ------------------------------------------------------------------------------------------------------

# Created on: 2018-30-6 10:41:50.00000
# Created by: Michael Wood
# Created for Robert Booth, GIS 6340  - GIS Customization Class - Northeastern University - Spring Qtr. 2018
# Copyright: 2018 - Michael Wood

# ------------------------------------------------------------------------------------------------------
# Overview:
# ------------------------------------------------------------------------------------------------------

# This Python script is an attempt to speed up an Esri ModelBuilder workflow that presently takes almost 24 hour to complete.

# Additionally this script will be able to be scheduled utilizing Microsoft Windows Task Manager to run overnight because it may take a long time to run to completion.

# ------------------------------------------------------------------------------------------------------
# Description: (What the Script Does)
# ------------------------------------------------------------------------------------------------------

# This python script automates a monotonous, tedious, and boring geoprocessing workflow.
# This geoprocessing workflow creates GIS feature classes from monthly text files that are provided by the United States Postal Service (USPS).
# In addition to being a very time-consuming process, it is also prone to human error during the processing of the GIS data
# due to a large number of geoprocessing operations. The script can also be adapted for other datasets and geographies like US counties.
# Inaddition, the script can be altered to be scheduled to run in Microsoft Task Scheduler to run over the weekend,

# The geoprocessing workflow proceeds in the following manner. Each month, an updated “Postal Carrier Route” (PCR) text 
# file is downloaded from the USPS.  A PCR is a sub-division of a ZIP Code and is generally the geography of the addresses a postal 
# carrier can deliver mail to in a workday. This file contains the number of total residences, counts of single and multi-family 
# households and businesses in each carrier route. In addition, the latitude and longitude of the route are also provided.  In addition, 
# a demographic data text file at ZIP Code level are provided thatwas obtained from US Census Bureau.   
# Lastly, there is a shapefile of the United States to be used for clipping operations and general reference purposes. 

# The script is comprised of primarily two distinct sections.  These are Pre-Geoprocessing Tasks (housekeeping functions)and Geoprocessing Operations.  In the 
# Pre-Geoprocessing tasks section, a Personal Geodatabase will be created, and multiple Feature Datasets will be created to segregate and organize 
# intermediate and "final output" feature classes.  The provided files will then be imported into the geodatabase so that the geoprocessing tools can 
# be run on the data.  The raw data is being imported and kept as it is a requirement to keep the original data with the resultant GIS 
# data. This data can also be utilized for other GIS related tasks or as a general reference.

# ----------------------------------------------
# Source Citation Section:
# ----------------------------------------------

# I would like to cite Professor Robert Booth for his GIS Customization Lecture Notes and class feedback as well as encourgament for making this script possible.
# Student participation and suggestions contibuted significantly as well.
# In addition to class lectures , materials and discussion, Environmental Systems Reasearch Institute (Esri) online resources such as forums and tutorials contributed
# signficantly to providing supplementary info thst made this script possible. http://resources.arcgis.com/en/communities/python/
# The book "Python Scripting for ArcGIS by Paul A. Zanderbergen contributed significantly to making this Python script possible
# Lasty, contributers to the website sourceforge provided additional insights on how to implement many arcpy functions as well as proper syntax use.

# ----------------------------------------------
# Data Source Citation Section:
# ----------------------------------------------

# US_State.shp: (2017). Unites States State TIGER Shapefile. US Census Bureau.
# Obtained from https://www.census.gov/cgi-bin/geo/shapefiles/index.php?year=2017&layergroup=States+%28and+equivalent%29

# USPS_PCR_Count_LL.txt. (2018). United States Postal Carrier Routes. United States Postal Service (USPS). Obtained from United States Postal Service.

# USPS_ZIP_Demographics.txt. (2017). ZIP Code Demographics. Obtained from https://factfinder.census.gov/faces/nav/jsf/pages/index.xhtml.


# ----------------------------------------------
# Important Acronyms Used For Variable Naming:
# ----------------------------------------------

# UPSP: United States Postal Service
# PCR: Acronym Stands for: Postal Carrier Routes 
# CRRT: Acronym Stands for: Postal Carrier Routes. This Acronym is used interchangably with PCR
# ZIP: Acronym Stands for: ZIP Codes. ZIP stands for "Zone Improvment Program"
# XY_Event: is the Process of converting a tabular data file containing Lattitude and Longitude coordinates to a Points Feature Class in ArcGIS.
# Geoprocessing: Geoprocessing is a framework and set of tools for processing geographic and related data.
# PGDB: Acronym Stands for: Personal Geodatabase. 

# ----------------------------------------------
# Important Background Notes About USPS Data:
# ----------------------------------------------

# ZIP Codes and carrier routes almost never cross state boundaries.  I believe there are only three exceptions to this rule.
# Carrier routes try to not cross county boundaries. However, this occurs less than 20% of the time.

# (PGDB Choosen for convienience and Easy Distribution - Will not Exceed DB Limitations of Size or Users)

# Data for this Script Project Only Has New England States to Facilite Testing and Speed Up Processing Time.

# This Model Takes Approx. 24 Hours to Process the Entire United States Dataset.

# ------------------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------------------

# Begininning of Script

# ------------------------------------------------------------------------------------------------------
# ------------------------------------------------------------------------------------------------------

# Import ArcPY and Other operating Specific Modules

import arcpy                        # This statement loads the Esri ArcGIS geoprocessing Tools & Utilities.
from arcpy import env               # This loads the "env" (Workspace Environment) clsss contained in the ArcPy Module.
import os                           # Loads the standard Python "os" (Operating System Module).
                                    # the OS module allows an interface with the underlying operating system.
import sys                          # Loads the "sys" module required to perform certain System - specific parameters and functions. 

# ------------------------------------------------------------------------------------------------------

# ArcPy Overwrite Existing Data Command

arcpy.env.overwriteOutput = True    # This overwites prexisting data unless there is a file lock in place possibly due to the file being open or in use by another program.

# ------------------------------------------------------------------------------------------------------
# Script arguments -*- GetParametersFromText() Section

# GetParametersFromText User Specified Input and Output Area
# GetParametersFromText() Order in Script Must Match Order in Script Tool
# ------------------------------------------------------------------------------------------------------

GISData = arcpy.GetParameterAsText(0)                                   # This allows a user to select the location to create the Personal Geodatabase Drive and Folder Location.
                                                                        # The folder location to create the Personal Geodatabase must already exist
                                                                        # Functionality to create a new folder will be added at a later time.

if GISData == '#' or not GISData:                                       # Default Value is C:\Data per company Standard Operating Proceedures provide a default value if unspecified
    GISData = "C:\\Temp"                                                # Required in Script Tool - Data Type: Folder - Direction: Input
                                                                        # Required in Script Tool - Data Type: Any value - Direction: Output
                                                    
Personal_GDB_Name = arcpy.GetParameterAsText(1)                         # This allows the user to specify the name of the Personal Geodatabase
                                                                        # (Remember .mdb File Extension - MS Access File Format - Has Limitations not Pertinant Here)
                                                
if Personal_GDB_Name == '#' or not Personal_GDB_Name:
    Personal_GDB_Name = "pgdb"                                          # provide a default value if unspecified

US_States_shp = arcpy.GetParameterAsText(2)                             # The Provided United States Shapefile from US Census Bureau" User Selected Input Data.
if US_States_shp == '#' or not US_States_shp:
    US_States_shp = "C:\\Temp\\US_States.shp"                           # provide a default value if unspecified

USPS_PCR_Counts_LL_txt = arcpy.GetParameterAsText(3)                    # The Provided United States Counts File with Lat/Long from US Postal Service" User Selected Input Data.
if USPS_PCR_Counts_LL_txt == '#' or not USPS_PCR_Counts_LL_txt:
    USPS_PCR_Counts_LL_txt = "C:\\Temp\\USPS_PCR_Counts_LL.txt"         # provide a default value if unspecified

USPS_ZIP_Demographics_txt = arcpy.GetParameterAsText(4)                 #The Provided United States ZIP Code Demographics File Derived From US Census Data" User Selected Input Data.
if USPS_ZIP_Demographics_txt == '#' or not USPS_ZIP_Demographics_txt:
    USPS_ZIP_Demographics_txt = "C:\\Temp\\USPS_ZIP_Demographics.txt"   # provide a default value if unspecified

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Local variables:
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
pgdb_mdb = GISData                                                      # Set the Personal Geodatabase Name Variable to GISData.
pgdb_ext = '.mdb'                                                       # Personal Geodatabase Extension (.mdb) and not (File Geodatabase Extension .gdb).

#This has to be Defined before othe Statements and Functions Can Be Run.
Postage_PGDB = GISData + "\\" + Personal_GDB_Name + pgdb_ext            # Personal Geodatabase Path Workaround for Variables.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

USStates = 'US_States'                                                                                                  # The US Census United States Shapefile (New England States Only for Demonstration Purposes).
XYEvents = 'XY_Events'                                                                                                  # Variable Name for Path of XY_Events Feature Dataset.
Geo_Processing = 'Geoprocessing'                                                                                        # Variable Name for Path of Geoprocessing Feature Dataset.
US_StatesFDS = 'US_States'                                                                                              # Variable Name for Path of US_States Feature Dataset.
PCRCentroids = 'PCR_Centroids'                                                                                          # Variable Name for Path of PCR_Centroids Feature Dataset.
PCRPolygons = 'PCR_Polygons'                                                                                            # Variable Name for Path of PCR_Polygonss Feature Dataset.
ZIPCentroids = 'ZIP_Centroids'                                                                                          # Variable Name for Path of ZIP_Centroids Feature Dataset.
ZIPPolygons = 'ZIP_Polygons'                                                                                            # Variable Name for Path of ZIP_Polygons Feature Dataset.
USPS_PCR_CountsLL = 'USPS_PCR_Counts_LL'                                                                                # Variable Name for Path of USPS Counts Text File.
PCRDemographics = 'PCR_Demographics'                                                                                    # Variable Name for Path of US Census Derived Carrier Routes (PCR)Demographics Text File.
ZIPDemographics = 'ZIP_Demographics'                                                                                    # Variable Name for Path of US Census Derived ZIP Code Demographics Text File.
USPS_XYZ_CountsLL_Layer = "USPS_XYZ_Counts_LL_Layer"                                                                    # XY_Events Process to Create XYZ_Events Layer.
USPS_XYZ_BaseData = "USPS_XYZ_Base_Data"                                                                                # Process to Copy XYZ_Events Layer to PGDB.
US_States = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + USStates                                             # Path to US_States Shapefile Variable Input.
XY_Events = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + XYEvents                                             # Path to XY_Events Feature Dataset File in PGDB.
Geoprocessing = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + Geo_Processing                                   # Path to Geo_processing Feature Dataset File in PGDB.
ZIPProcessing = 'ZIP_Processing'                                                                                        # Variable for creating the ZIP_Processing Feature Dataset
ZIP_Processing = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPProcessing                                   # Path to ZIP_Processing Feature Dataset File in PGDB.
ScratchFDS = 'Scratch'                                                                                                  # Variable for for creating the Scratch Feature Dataset
Scratch = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ScratchFDS                                             # Path to Scratch Feature Dataset File in PGDB.
USPS_PCR_Counts_LL = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + USPS_PCR_CountsLL                           # Path to USPS_PCR_Counts_LL Text File in PGDB.
USPS_XYZ_Counts_LL_Layer = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + USPS_XYZ_CountsLL_Layer               # Path to USPS_PCR_CountsLL Layer File in PGDB.
USPS_XYZ_Base_Data = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + XYEvents + "\\" + USPS_XYZ_BaseData         # Path to Thiessen Polygons Feature Dataset File in PGDB (XY_Events FDS).
US_States_FDS = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + US_StatesFDS                                     # Path to US_States Feature Dataset File in PGDB.
PCR_Centroids = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRCentroids                                     # Path to PCR_Centroids Feature Dataset File in PGDB.
PCR_Polygons = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRPolygons                                       # Path to PCR_Polygons Feature Dataset File in PGDB. 
ZIP_Centroids = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPCentroids                                     # Path to ZIP_Centroids Feature Dataset File in PGDB. 
ZIP_Polygons = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons                                       # Path to ZIP_Polygons Feature Dataset File in PGDB. 
PCR_Demographics = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRDemographics                               # Path to PCR_Demogrpahics Text File Dataset File in PGDB.
ZIP_Demographics = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPDemographics                               # Path to ZIP_Demogrpahics Text File Dataset File in PGDB.
USPS_XYZ_Counts_LL_wDemo_Layer = USPS_XYZ_Counts_LL_Layer                                                               # XYZ_Events and Join with Demographic Data EquivalencyThe Join
PCR_Centroids_wDemoNE = 'PCR_Centoids_wDemo_NE'                                                                         # Variable for Path for saving ZIP Code Centroids with Demographics (For New England States Only!)
PCR_Centroids_wDemo_NE = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + XYEvents + "\\" + PCR_Centroids_wDemoNE # Path for saving ZIP Code polygons with Demographics (For New England States Only!)


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Pre-Geoproessing Section (Housekeeping)
# Geodatabase Creation, Import and Data/Geoprocessing Statemtns and Functions.

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Personal Geodatabase Creation (With Script Tool - User Defined Locations).
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Create Personal GDB
# This creates a Personal Geodatabase (PGDB)

arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("***** Begin Step 1 *****")                                                    #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("Creating Personal Geodatabase")                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# arcpy.env.workspace = "C:\\Users\\Administrator\\Documents\\ArcGIS\\Default.gdb"              # Set the workspace environment (directory)

# This uses the ESRI geoprocessing tool to create a Personal Geodatabase (PGDB)
# Using input for GetParametersFromText(0) and GetParametersFromText(1) in the Script Tool

arcpy.CreatePersonalGDB_management(GISData, Personal_GDB_Name, "CURRENT")                       # Run Esri CreatePersonalGDB Tool using default values 

arcpy.AddMessage("Personal Geodatabase Created Succesfully")                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("***** End Step 1 *****")                                                      #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data Import US_States Shapefile into Personal Geodatabase.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Copy Features US States Shapefile

arcpy.AddMessage("***** Begin Step 2 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Copying US States Shapefile")                                                 #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment (Directory) tor the root of the PGDB

# Copy the US_States.shp shapefile from the location specified in GetParametersFromText(2)

arcpy.CopyFeatures_management(US_States_shp, US_States, "", "0", "0", "0")                      # Run Esri CopyFeatures Tool with default values 

#~~~~~~~~~~~~~
#Get the count from GetCount's Result object
#~~~~~~~~~~~~~

feature_count = int(arcpy.GetCount_management(US_States).getOutput(0))                          # Count the numer of states in the US_States Feature in the PGDB

if feature_count == 0:                                                                          # If statement to determine if feature has no features 
    arcpy.AddError("{0} has no features.".format(US_States))                                    # Return error if no features
else:                                                                                           # Otherwise
    arcpy.AddMessage("{0} has {1} features.".format(US_States, feature_count))                  # List the number of features (US States) in feature US_States.
    arcpy.AddMessage("  ")                                                                      # Blank space spacer

    arcpy.AddMessage("US States Shapefile Copied Succesfully")                                  #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    arcpy.AddMessage("***** End Step 2 *****")                                                  #   End Tool Message Dispalyed in Script Tool
    arcpy.AddMessage("  ")                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
         
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#Create "Intermediate Data" Feature Datasets to facilitate Gorprocessing Operations
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Note: These "Intermediate Data" Feature Datasets will be deleted at the completion of the script

arcpy.AddMessage("***** Begin Step 3 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Creating Intermediate Data Feature Datasets")                                 #   Begin Tool Message Dispalyed in Script Tool

# Process: Create XY_Events Feature Dataset

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CreateFeatureDataset_management(Postage_PGDB, "XY_Events", "")                            # Feature Dataset created to store US PCR Centroids to be Clipped into 
                                                                                                # individual state PCR Centroids to be placed into the PCR_Centroids Feature Dataset

# Process: Create Geoprocessing Feature Dataset

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CreateFeatureDataset_management(Postage_PGDB, "Geoprocessing", "")                        # Feature Dataset created to store US Thiessen Polygons to be Clipped into
                                                                                                # individual state PCR Polygons to be placed into the PCR_Polygons Feature Dataset

# Process: Create ZIP Geoprocessing Feature Dataset

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CreateFeatureDataset_management(Postage_PGDB, "ZIP_Processing", "")                       # Feature Dataset created to store Disolved Thiessen Polygons to be Clipped into
                                                                                                # individual state ZIP Polygons to be placed into the ZIP_Polygons Feature Dataset
# Process: Create Scratch Feature Dataset

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CreateFeatureDataset_management(Postage_PGDB, "Scratch", "")                              # Feature Dataset created to store ZIP Centroids to be Clipped into
                                                                                                # individual state ZIP to be placed into the ZIP_Centroids Feature Dataset
                                                                                                
arcpy.AddMessage("Intermediate Data Feature Datasets Created")
arcpy.AddMessage("***** End Step 3 *****")                                                      #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Data Import USPS PCR Counts text file containing Lat/Long (XY) information into Personal Geodatabase.
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Copy USPS Lat Long Rows

arcpy.AddMessage("***** Begin Step 4 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Copying USPS Counts File to PGDB")                                            #   Begin Tool Message Dispalyed in Script Tool

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CopyRows_management(USPS_PCR_Counts_LL_txt, USPS_PCR_Counts_LL, "")                       # Run Esri CopyRows tool to import the USPS Data 

arcpy.AddMessage("USPS Counts File Copied Succesfully to PGDB")
arcpy.AddMessage("***** End Step 4 *****")                                                      #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create the USPS PCR XY Event Layer in Personal Geodatabase 
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Make XY Event Layer

arcpy.AddMessage("***** Begin Step 5 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Making USPS Counts XY Event Layer")                                           #   Begin Tool Message Dispalyed in Script Tool

# This step creates a XY Event Layer usinging Lat/Long from the USPS Latitude and Longitude values
# This uses the Esri MakeXYEventLayer tool using default values
# and is used to place results into XY_Events Feature Dataset.

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.MakeXYEventLayer_management(USPS_PCR_Counts_LL, "lon", "lat", USPS_XYZ_Counts_LL_Layer, "GEOGCS['GCS_WGS_1984',DATUM['D_WGS_1984',SPHEROID['WGS_1984',6378137.0,298.257223563]],PRIMEM['Greenwich',0.0],UNIT['Degree',0.0174532925199433]];-400 -400 1000000000;-100000 10000;-100000 10000;8.98315284119522E-09;0.001;0.001;IsHighPrecision", "BOXCNT")

arcpy.AddMessage("USPS Counts XY Event Layer Cereated")
arcpy.AddMessage("***** End Step 5 *****")                                                      #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Process: Copy XYZ_Event_Features to Feature Dataset XY_Events in PGDB
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 6 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Copy out USPS Counts Lat/Long Layer to PGDB")                                 #   Begin Tool Message Dispalyed in Script Tool

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB

# This will write the results of the XY_Event Layer to the XY_Events Feature Dataset for further geoprocessing processing operations
# This dataset will be the foundation for all subsequent geoprocessing operations

arcpy.CopyFeatures_management(USPS_XYZ_Counts_LL_Layer, USPS_XYZ_Base_Data, "", "0", "0", "0")

arcpy.AddMessage("USPS Counts Lat/Long Layer Has Been Copied to PGDB")
arcpy.AddMessage("***** End Step 6 *****")                                                      #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Personal Geodatabase "Final Results" Feature Dataset Creation Section
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 7 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Creating Output Data Feature Datasets")                                       # Begin Tool Message Dispalyed in Script Tool

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB

# Process: Create US_States Feature Dataset

arcpy.CreateFeatureDataset_management(Postage_PGDB, "US_States", "")                            # Creates US_States final output Feature Dataset

# Process: Create PCR_Centroids Feature Dataset

arcpy.CreateFeatureDataset_management(Postage_PGDB, "PCR_Centroids", "")                        # Creates PCR_Centroids final output Feature Dataset

# Process: Create PCR Polygons Feature Dataset

arcpy.CreateFeatureDataset_management(Postage_PGDB, "PCR_Polygons", "")                         # Creates PCR_Polygons final output Feature Dataset

# Process: Create ZIP_Centroids Feature Dataset

arcpy.CreateFeatureDataset_management(Postage_PGDB, "ZIP_Centroids", "")                        # Creates ZIP_Centroids final output Feature Dataset

# Process: Create ZIP_Polygons Feature Dataset

arcpy.CreateFeatureDataset_management(Postage_PGDB, "ZIP_Polygons", "")                         # Creates ZIP_Polygons final output Feature Dataset

arcpy.AddMessage("Output Feature Datasets Created")
arcpy.AddMessage("***** End Step 7 *****")                                                      #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~     

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Process: Copy ZIP Code Demographic Rows
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 8 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Creating Output Data Feature Datasets")                                       # Begin Tool Message Dispalyed in Script Tool

# This step copies the US Census Bureau Demographic Data text file to the root of the PGDB
# This file will be joined to the final ZIP Code Polygonas and Centroid features
# to provide additional information for GIS data processing

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB
arcpy.CopyRows_management(USPS_ZIP_Demographics_txt, ZIP_Demographics, "")

arcpy.AddMessage("Output Feature Datasets Created")
arcpy.AddMessage("***** End Step 8 *****")                                                      #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#US_States to US_States Feature Dataset (Individual States Clip Section)
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 9 *****")                                                    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Selecting US States in Shapefile to Create")                                  # Begin Tool Message Dispalyed in Script Tool 
arcpy.AddMessage("Individual US States Shapefiles in PGDB")                                     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = Personal_GDB_Name                                                         # Set the environment to the root of the PGDB

# Use a SearchCursor to select each state in the attribute table of the US_States feature in the PGDB

rows = arcpy.SearchCursor(US_States)
for row in rows:
    feat = row.Shape
    arcpy.Select_analysis(feat, US_States_FDS + "\\" + str('State_of_') + str(row.getValue('ST_ABBRV')), '')

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + USStates + "\\"

# Count the states in the US_States Feature Dataset

for fc in arcpy.ListFeatureClasses():
        name = os.path.splitext(os.path.basename(fc))[0]

        #Get the count from GetCount's Result object

        feature_count = int(arcpy.GetCount_management(fc).getOutput(0))

        if feature_count == 0:
            
             arcpy.AddError("{0} has no features.".format(fc))
        else:
             arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))

arcpy.AddMessage("  ")                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Individual US States Shapefiles")                                         #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("have been Created in PGDB")                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("***** End Step 9 *****")                                                  # End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# PCR Centroids to PCR_Centroids Feature Dataset (PCRs to Individual states Clip)
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 10 *****")                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Clipping XY Event Layer into")                                            # Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("PCR Individual State Centroids")                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Please Wait This Takes a While")                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = Personal_GDB_Name                                                     # Set the environment to the root of the PGDB

# Use a SearchCursor to select each state in the US_States Feature in the PGDB
# to be used to clip the PCR XY_Event layer into individual state PCR Features
# The name of the state each individual state will be PCR_Centroids_ and the state abbreviation (ie PCR_Centroids__NY)

rows = arcpy.SearchCursor(US_States)
for row in rows:
    feat = row.Shape
    arcpy.Clip_analysis(USPS_XYZ_Base_Data, feat, PCR_Centroids + "\\" + str('PCR_Centroids__') + str(row.getValue('ST_ABBRV')), '')

# In this section a field is being added to the PCR_Centroids so that the ZIP Code can be converted to a text value
#  so it will not lose the leading zero that occures in the ZIP Codes of the New England States. (Thank you so much USPS)
# To accomplish this, the Esri Add field and Calculate field tools are employed

    #Process: Add Field

    arcpy.AddField_management(PCR_Centroids + "\\" + str('PCR_Centroids__') + str(row.getValue('ST_ABBRV')), "ZIPCODE_T", "TEXT", "", "", "5", "", "NULLABLE", "NON_REQUIRED", "")

    #Process: Calculate Field

    arcpy.CalculateField_management(PCR_Centroids + "\\" + str('PCR_Centroids__') + str(row.getValue('ST_ABBRV')), "ZIPCODE_T", "\"0\" + str(!ZIP!)", "PYTHON_9.3", "")

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRCentroids + "\\"    # Set the environment to the PCR_Centroids Feature Dataset to get count

# Use Esri ListFeatureClasses to make list of all states in PCR_Centroids Feature Dataset to obtain Count of Features in each state 

for fc in arcpy.ListFeatureClasses():
        name = os.path.splitext(os.path.basename(fc))[0]

        #Get the count from GetCount's Result object

        feature_count = int(arcpy.GetCount_management(fc).getOutput(0))                                     # Counts the number of Centroids in each state to display in Script Tool Window.

        if feature_count == 0:                                                                              # If statement to determine if feature has no features and return error    
            
             arcpy.AddError("{0} has no features.".format(fc))                                              # If there are features, it should dispaly error message
        else:                                                                                               # Otherwise
             arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))                            # List the number of features (USPS Carrier Routes) in each state in the feature dataset.

arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~         
arcpy.AddMessage("XY Event Layer Clipped into")                                                             #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage(" PCR Individual State Centroids")                                                         #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("***** End Step 10 *****")                                                                 # End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Create Thiessen Polygons from Shapefiles in Feature Dataset PCR_Polygons (Individual Thiessen Polygons)
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 11 *****")                                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Creating Thiessen Polygons")                                                              #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Please Wait, This Takes a Moment")                                                        #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRCentroids + "\\"            # Set the environment (directory) to PCR Centroids feature dataset. PCR = Postal Carrier Route

Thiessen = 'Thiessen_Polygons_'

for fc in arcpy.ListFeatureClasses():                                                                       # Create List of Features (States) in PCR Centroids feature dataset
        name = os.path.splitext(os.path.basename(fc))[0]                                                    # This will obtain the name of each feature in the feature dataset (ie. New_York_NY).
        id = name[-2:]                                                                                      # This will obtain the state abbreviation at the end of each statename to be used as a variable later. (ie. NY)
        print id                                                                                            # Print the state abbreviation.(NY)

        # Run the Esri geoprocessing tool to Create Thiessen Polygons.
        # in the Geo_Processing Feature Dataset in the Personal Geodatabase
        # The name of the state each individual state will be Thiessen_Polygons_ and the state abbreviation (ie Thiessen_Polygons_NY)
        
        arcpy.CreateThiessenPolygons_analysis(fc, GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + Geo_Processing + "\\" + 'Thiessen_Polygons_' + str(id),'ALL') 

        #Get the count from GetCount's Result object

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + Geo_Processing + "\\"  # Set the environment (directory) to Temporary Data Holder for Thiessen Polygons called Geo_Processing

for fc in arcpy.ListFeatureClasses(): 

    feature_count = int(arcpy.GetCount_management(fc).getOutput(0))                                         # Counts the number of Thiessen Polygons in each state to display in Script Tool Window.

    if feature_count == 0:                                                                                  # If statement to determine if feature has no features and return error                                                                           
            
        arcpy.AddError("{0} has no features.".format(fc))                                                   # If there are features, it should dispaly error message 
    else:                                                                                                   # Otherwise
        arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))                                 # List the number of features (USPS Carrier Routes) in each state in the feature dataset.

#del cursor

arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Intermediate Thiessen Polygons Created")                                                  #       End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("***** End Step 11 *****")                                                                 #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Clip PCR Polygons to Individual States
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Here I found code on the Internet that assisted me in writing my clipping operation
# I am unable to find the source where I orginally copied the sample script
# As I am unable to properly cite the author of the sample code I emulated
# I am citing the fact that this code is not entirely my own
# Thank you whoever you are, you helped me solve a big problem.

# What this helped solve was how to iterate through a multipart feature (US_States) to select each state
# clip an individual features (individual states) in in a feature dataset 

arcpy.AddMessage("***** Begin Step 12 *****")                                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Clipping Thiessen Polygons to PCR Polygons")                                              #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("Please Wait, This Takes a Moment")                                                        #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = Personal_GDB_Name                                                                     # Set the environment (directory) to the root of the Personal Geodatabase

# Get the value from field "SO_NAME" which is State_of_ from each feature (state) in the US_States feature class
# for building the clip features path

field = "SO_NAME"                                                                                       
cursor = arcpy.SearchCursor(US_States)                                                                  
for row in cursor:
    feat2 = (row.getValue(field))                                                                       
    
env.workspace = GISData + '\\' + Personal_GDB_Name + pgdb_ext + '\\' + Geo_Processing + '\\'                # Set the environment (directory) the intermediate Thiessen polygons

path2 = GISData + '\\' + Personal_GDB_Name + pgdb_ext + '\\' + USStates + '\\'                              # Set path2 variable to to US_States features for building the clip features path
path3 = GISData + '\\' + Personal_GDB_Name + pgdb_ext + '\\' + PCRPolygons + '\\'                           # Set path3 variable to to PCR_Polygons feature dataset for building the clip features path
files = arcpy.ListFeatureClasses()                                                                          # Set files variable to arcpy.ListFeatureClasses() tool
print files

for fc in files:                                                                                            # Create List of Thiessen polygons in Intermediate Geoprocessing Feature Dataset
        name = os.path.splitext(os.path.basename(fc))[0]
        id = name[-2:]                                                                                      # Select last two characters of each feature which is the state abbreviation

        # Run Esri Clip_analysis toolto clip Thiessen polygons in the Geoprocessing Feature Dataset
        # using the US_States features
        # This will create final output PCR Polygons as individual states in the PCR_Polygons Feature Dataset

        arcpy.Clip_analysis(fc, path2 + feat2 + (id), path3  +  'PCR_Polygons__' + str(id), '')
        
        # Add a field to thenewly created PCR polygons to convert numbers to text
        # this is done to add leading zeros to New England State ZIP Codes.

        #Process: Add Field

        arcpy.AddField_management(path3  +  'PCR_Polygons__' + str(id), "ZIPCODE_T", "TEXT", "", "", "5", "", "NULLABLE", "NON_REQUIRED", "")

        #Process: Calculate Field

        arcpy.CalculateField_management(path3  +  'PCR_Polygons__' + str(id), "ZIPCODE_T", "\"0\" + str(!ZIP!)", "PYTHON_9.3", "")  # Calculate the ZIP Code for the new field and add the leading zero

#Get the count from GetCount's Result object

arcpy.env.workspace = GISData + '\\' + Personal_GDB_Name + pgdb_ext + '\\' + PCRPolygons + '\\'             # Set the environment (directory) to Permanant Feature Dataset called PCR_Polygons

for fc in arcpy.ListFeatureClasses():                                                                       # Create a list of features in Feature Dataset called PCR_Centroids

    feature_count = int(arcpy.GetCount_management(fc).getOutput(0))                                         # Counts the number of Point Centroids in each state to display in Script Tool Window.

    if feature_count == 0:                                                                                  # If statement to determine if feature has no features and return error                                                                           
            
        arcpy.AddError("{0} has no features.".format(fc))                                                   # If there are features, it should dispaly error message 
    else:                                                                                                   # Otherwise
        arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))                                 # List the number of features  


arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~    
arcpy.AddMessage("Clipped Thiessen Polygons to PCR Polygons")                                               #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("***** End Step 12 *****")                                                                 #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#Dissolve Section Carrier Route Polygons to ZIP Code Polygons
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 13 *****")                                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Dissolving PCR Polygons to Created ZIP Code Polygons")                                    #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + PCRPolygons + "\\"             # Set environment (directory)to PCR_Polygons Feature Dataset in PGDB 

for fc in arcpy.ListFeatureClasses():                                                                       # Create list of states in CR_Polygons Feature Dataset in PGDB
        name = os.path.splitext(os.path.basename(fc))[0]                                                    # This will obtain the name of each feature in the feature
        id = name[-2:]                                                                                      # This will obtain the state abbreviation at the end of each statename to be used as a variable later. (ie. NY)                                                                      
        print id                                                                                            # Print the state abbreviation.(NY)

        # This step uses the Esri Dissolve_management tool to dissolve the PCR Polygons
        # To create a derivitive dataset which are USPS ZIP Codes
        # The intermediate data is saved to the ZIP_Processing intermediate Feature Dataset
        
        arcpy.Dissolve_management(fc, GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPProcessing + "\\" +  'Dissolve_Polygons_' + str(id), "ZIP", "ZIPCRID COUNT;TOTRESCNT SUM;SFDU SUM;MFDU SUM;BOXCNT SUM;BIZCNT SUM", "MULTI_PART", "DISSOLVE_LINES")

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPProcessing + "\\"           # Set environment (directory)to ZIP_Processing Feature Dataset in PGDB        

#Get the count from GetCount's Result object

for fc in arcpy.ListFeatureClasses():                                                                       # Create list of states in ZIP_Processing Feature Dataset in PGDB  

    feature_count = int(arcpy.GetCount_management(fc).getOutput(0))                                         # Count the number of dissolve features in each state in the Feature Dataset in the PGDB

    if feature_count == 0:                                                                                  # If statement to determine if feature has no features 
        
        arcpy.AddError("{0} has no features.".format(fc))                                                   # Return error if no features
    else:                                                                                                   # Otherwise
        arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))                                 # List the number of features (Zip COdes) in Feature Dataset

arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("PCR Polygons Have Dissolved to Create ZIP Code Polygons")                                 #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("***** End Step 13 *****")                                                                 #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#del cursor

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ZIP Code - JOIN Section 
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 14 *****")                                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Joining US Census Demographic Data to ZIP Code Polygons")                                 #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("and Creating ZIP COde Polygons")                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Make Feature Layer & Join & Copy Data Out to ZIP Code Polygons ZIP_Polygons_(State Abberviation)

Dissolve_Polygons_Layer = 'Dissolve_Polygons_Layer_'                                                    

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPProcessing + "\\"           # Set Environment (Directory) to the Intermediate ZIP_Processing Feature Dataset

for fc in arcpy.ListFeatureClasses():
        name = os.path.splitext(os.path.basename(fc))[0]
        id = name[-2:]
        print id

        # This uses the Esri MakeFeatureLayer tool to create intermediate features
        # from the dissloved PCR polygons
            
        arcpy.MakeFeatureLayer_management(fc, Dissolve_Polygons_Layer, "", "", "OBJECTID OBJECTID VISIBLE NONE;ZIP ZIP VISIBLE NONE")

        # This uses the Esri AddJoin_management tool to
        # Join the US Census data to the dissolved polgons (ZIP Codes)to provide additional information for GIS analysis
        # The join fields are ZIP and KEY Postal in the dissolved polygons and ZIP_Demographics table respectivly
                        
        arcpy.AddJoin_management(Dissolve_Polygons_Layer, "ZIP", ZIP_Demographics, "KEY_POSTAL", "KEEP_COMMON")

        # This copies the results of the join to the ZIP_Polygons Final Results ZIP_Polygons Feature Dataset
       
        arcpy.CopyFeatures_management(Dissolve_Polygons_Layer, GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons + "\\" "ZIP_Code_Polygons__" + str(id))

        # Process: Add Field

        # This adds a field to the resulatant ZIP Code polygons to convert numbers to text.
        # This is neccisary so that so that the New England States will have "leading zeros"
        # Thank you New England for having leading zeros in your ZIP COdes

        arcpy.AddField_management(GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons + "\\" "ZIP_Code_Polygons__" + str(id), "ZIPCODE_T", "TEXT", "", "", "5", "", "NULLABLE", "NON_REQUIRED", "")

        # Process: Calculate Field

        #This calculates the "leading zeros" for the New England States and populates the "ZIPCODE_T" field

        arcpy.CalculateField_management(GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons + "\\" "ZIP_Code_Polygons__" + str(id), "ZIPCODE_T", "\"0\" + str(!ZIP_Demographics_KEY_POSTAL!)", "PYTHON_9.3", "")

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons + "\\"             # Set Environment (directory) to the ZIP_Polygons Final Output Feature Dataset

        #Get the count from GetCount's Result object

for fc in arcpy.ListFeatureClasses():

    feature_count = int(arcpy.GetCount_management(fc).getOutput(0))

    if feature_count == 0:
            
        arcpy.AddError("{0} has no features.".format(fc))
    else:
        arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))
             
del cursor

arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("US Census Demographic Data Has Been Joined to ZIP Code Polygons")                         #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("and Create ZIP Code Polygons")                                                            #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("***** End Step 14 *****")                                                                 #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#ZIP Code Polygons to ZIP Code Centroid Section 
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 15 *****")                                                               #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Creating ZIP Code Centroids with Deomgraphic Data")                                       #   Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                                      #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Make Feature Layer & Join & Copy Data Out to ZIP Code Polygons ZIP_Polygons_(State Abberviation)

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPPolygons + "\\"             # Set Environment (directory) to the ZIP_Polygons Final Output Feature Dataset

for fc in arcpy.ListFeatureClasses():                                                                       # Create list of states in ZIP_Polygons Feature Dataset in PGDB 
        
        name = os.path.splitext(os.path.basename(fc))[0]
        id = name[-2:]
        print id

        # This uses the Esri FeatureToPoint_management tool to
        # Convert indivdual ZIP Code polygons to individual ZIP Code Centroids
        # The results are placed in the ZIP_Centroids final output Feature Dataset in the PGDB
        # With unique file names appended to the end of each state.

        arcpy.FeatureToPoint_management(fc,GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPCentroids + "\\" + 'ZIP_Code_Centroids__' + str(id), 'CENTROID')

arcpy.env.workspace = GISData + "\\" + Personal_GDB_Name + pgdb_ext + "\\" + ZIPCentroids + "\\"                # Set Environment (directory) to the ZIP_Centroids Final Output Feature Dataset            

        #Get the count from GetCount's Result object

for fc in arcpy.ListFeatureClasses():

    feature_count = int(arcpy.GetCount_management(fc).getOutput(0))

    if feature_count == 0:
            
        arcpy.AddError("{0} has no features.".format(fc))
    else:
        arcpy.AddMessage("{0} has {1} features.".format(fc, feature_count))

arcpy.AddMessage("  ")                                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~      
arcpy.AddMessage("Created ZIP Code Centroids With Deomgraphic Data")                                            #   End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("***** End Step 15 *****")                                                                     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#del cursor

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# Delete Intermediate Data Section 
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("***** Begin Step 16 *****")                                                                   #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Deleteing Intermediate Data and Feature Datasets")                                            # Begin Tool Message Dispalyed in Script Tool
arcpy.AddMessage("  ")                                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

# Process: Delete

# Set workspace
arcpy.env.workspace = Personal_GDB_Name                                                                         # Set the workspace (directory) to the root of the Personal Geodatabase

# Set Local Variables
del_dir1 =  XY_Events                                                                                           # Set variable to delete intermendiate XY Event layer
del_dir2 = Scratch                                                                                              # Set variable to delete intermendiate ZIP Centroids                  
del_dir3 = Geoprocessing                                                                                        # Set variable to delete intermendiate Thiessen polygons                                                                                
del_dir4 = ZIP_Processing                                                                                       # Set variable to delete intermendiate Dissolve polygons

# Execute Delete
arcpy.Delete_management(del_dir1)                                                                              # Delete intermendiate XY Event layer Feature Dataset
arcpy.Delete_management(del_dir2)                                                                              # Delete intermendiate ZIP Centroids Feature Dataset
arcpy.Delete_management(del_dir3)                                                                              # Delete intermendiate Thiessen polygons Feature Dataset 
arcpy.Delete_management(del_dir4)                                                                              # Delete intermendiate Dissolve polygons Feature Dataset

arcpy.AddMessage("  ")                                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("Intermediate Data and Feature Datasets Have Been Deleted")                                    # End Tool Message Dispalyed in Script Tool
arcpy.AddMessage("***** End Step 16 *****")                                                                     #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("  ")                                                                                          #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
arcpy.AddMessage("End Script")
arcpy.AddMessage("  ")
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

arcpy.AddMessage("Class is Finaly Over - Snoopy Dance :)")                                                      # THE END! I hope you like my script
arcpy.AddMessage("  ")
arcpy.AddMessage("  ")
arcpy.AddMessage("Created by: Michael J. Wood")                                                                 # Print Citation in Script Tool Window
arcpy.AddMessage("Coptright: JUNE 2018")                                                                        # Print copywrite notice and date in Script Tool Window
arcpy.AddMessage("  ")
arcpy.AddMessage("  ")